/* $Id: wait.c 533 2007-12-22 15:25:42Z lukasz $ */
/*
 *  FedStage DRMAA for PBS Pro
 *  Copyright (C) 2006-2007  Fedstage Systems Inc.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file wait.c
 * PBS specific waiting routines and exit_status handling functions.
 */

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <assert.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <drmaa_utils/attrib.h>
#include <drmaa_utils/compat.h>
#include <drmaa_utils/drmaa_impl.h>
#include <drmaa_utils/util.h>
#include <pbs_drmaa/pbs_drmaa.h>

#ifndef lint
static char rcsid[]
#	ifdef __GNUC__
		__attribute__ ((unused))
#	endif
	= "$Id: wait.c 533 2007-12-22 15:25:42Z lukasz $";
#endif


void
drmaa_wait_for_next_check(
		drmaa_session_t *session,
		drmaa_cond_t *cond,
		drmaa_mutex_t *mutex,
		const struct timespec *timeout,
		drmaa_err_ctx_t *err
		)
{
	struct timespec ts, *next_check = &ts;
	bool status_changed;

	if( timeout )
		DEBUG(( "-> drmaa_wait_for_next_check( timeout=%ld.%09ld )",
					timeout->tv_sec, timeout->tv_nsec ));
	else
		DEBUG(( "-> drmaa_wait_for_next_check( timeout=(null) )" ));
	drmaa_get_time( next_check, err );
	drmaa_ts_add( next_check, &session->pool_delay );
	if( timeout  &&  drmaa_ts_cmp( timeout, next_check ) < 0 )
		next_check = (struct timespec*)timeout;
	DEBUG(( "drmaa_wait_for_next_check: waiting untill %ld.%09ld",
				next_check->tv_sec, next_check->tv_nsec ));
	status_changed = drmaa_cond_timedwait( cond, mutex, next_check, err );
	if( !status_changed  &&  next_check == timeout )
		RAISE_DRMAA_1( DRMAA_ERRNO_EXIT_TIMEOUT );

	DEBUG(( "<- drmaa_wait_for_next_check; "
				"err=%d, next_check=%ld.%09ld, status_changed=%d",
				err->rc, next_check->tv_sec, next_check->tv_nsec,
				(int)status_changed
				));
}


void
drmaa_wait_for_single_job(
		drmaa_session_t *session,
		const char *job_id,
		int *status, drmaa_attr_values_t **rusage,
		bool dispose,
		const struct timespec *timeout,
		drmaa_err_ctx_t *err
		)
{
	drmaa_job_t *job = NULL;
	drmaa_job_set_t *set = session->jobs;

	DEBUG(( "-> drmaa_wait_for_single_job( job_id=%s )", job_id ));
	if( OK(err) )
		job = drmaa_job_find( set, job_id, err );
	if( OK(err)  &&  !session->with_wait_thread )
		drmaa_job_update_status( job, err );

	while( OK(err)  &&  !session->end  &&  job->status < DRMAA_PS_DONE )
	 {
		bool status_changed;
		DEBUG(( "drmaa_wait_for_single_job: waiting for %s to terminate",
					job->job_id ));
		if( session->with_wait_thread )
		 {
			if( timeout )
				status_changed = drmaa_cond_timedwait(
						&job->status_cond, &job->mutex, timeout, err );
			else
			 {
				drmaa_cond_wait( &job->status_cond, &job->mutex, err );
				status_changed = true;
			 }
			if( OK(err)  &&  !status_changed )
				RAISE_DRMAA_1( DRMAA_ERRNO_EXIT_TIMEOUT );
		 }
		else
		 {
			drmaa_wait_for_next_check( session,
					&job->status_cond, &job->mutex, timeout, err );
		 }
		DEBUG(( "drmaa_wait_for_single_job: woken up" ));

		if( OK(err)  &&  !session->with_wait_thread )
			drmaa_job_update_status( job, err );
	 }

	if( session->end )
		RAISE_DRMAA_1( DRMAA_ERRNO_EXIT_TIMEOUT );

	if( OK(err) )
		drmaa_job_get_termination_state(
				job, status, rusage, err );

	if( job )
	 {
		if( OK(err) && dispose )
		 {
			job->flags |= DRMAA_JOB_DISPOSED;
			/*
			 * drmaa_job_set_remove will release
			 * one reference (and remove one lock).
			 */
			drmaa_mutex_lock( &job->mutex, err );
			drmaa_job_set_remove( set, job, err );
		 }
		drmaa_job_release( &job, err );
	 }

	DEBUG(( "<- drmaa_wait_for_single_job" ));
}


char *
drmaa_wait_for_any_job(
		drmaa_session_t *session,
		int *status, drmaa_attr_values_t **rusage,
		bool dispose,
		const struct timespec *timeout,
		drmaa_err_ctx_t *err
		)
{
	drmaa_job_t *job = NULL;
	drmaa_job_set_t *set = session->jobs;
	char *job_id = NULL;

	DEBUG(( "-> drmaa_wait_for_any_job" ));

	if( OK(err)  &&  !session->with_wait_thread )
		drmaa_update_all_jobs_status( session, err );
	if( OK(err) )
		job = drmaa_job_find_terminated( set, err );

	while( OK(err)  &&  job == NULL )
	 {
		bool signaled = true;

		drmaa_mutex_lock( &session->end_mutex, err );
		if( session->end )
			RAISE_DRMAA_1( DRMAA_ERRNO_NO_ACTIVE_SESSION );
		if( OK(err) )
		 {
			if( session->with_wait_thread )
			 {
				DEBUG(( "drmaa_wait_for_any_job: waiting for wait thread" ));
				if( timeout )
					signaled = drmaa_cond_timedwait( &session->wait_cond, &session->end_mutex,
							timeout, err );
				else
					drmaa_cond_wait( &session->wait_cond, &session->end_mutex, err );
			 }
			else
			 {
				DEBUG(( "drmaa_wait_for_any_job: waiting for next check" ));
				if( OK(err) )
					drmaa_wait_for_next_check( session,
							&session->wait_cond, &session->end_mutex, timeout, err );
			 }
		 }
		drmaa_mutex_unlock( &session->end_mutex, err );
		DEBUG(( "drmaa_update_all_jobs_status: woken up; signaled=%d; err=%d",
					signaled, err->rc ));

		if( OK(err) && !signaled )
		 {
			RAISE_DRMAA_1( DRMAA_ERRNO_EXIT_TIMEOUT );
			break;
		 }

		if( OK(err)  &&  !session->with_wait_thread )
			drmaa_update_all_jobs_status( session, err );
		if( OK(err) )
			job = drmaa_job_find_terminated( set, err );
	 }
	DEBUG(( "drmaa_wait_for_any_job: waiting finished" ));

	if( OK(err) )
		job_id = drmaa_strdup( job->job_id, err );
	if( OK(err) )
		drmaa_job_get_termination_state(
				job, status, rusage, err );
	if( job )
	 {
		if( OK(err) && dispose )
		 {
			/*
			 * drmaa_job_set_remove will release
			 * one reference (and remove one lock).
			 */
			drmaa_mutex_lock( &job->mutex, err );
			drmaa_job_set_remove( set, job, err );
		 }
		drmaa_job_release( &job, err );
	 }

	if( !OK(err) && job_id )
	 {
		DRMAA_FREE( job_id );
		job_id = NULL;
	 }
	return job_id;
}


void *
drmaa_wait_thread( void *arg )
{
	drmaa_session_t *session = (drmaa_session_t*)arg;
	drmaa_err_ctx_t errctx, *err=&errctx;

	DEBUG(( "-> drmaa_wait_thread" ));
	drmaa_err_create( err, NULL );
	while( OK(err) )
	 {
		drmaa_update_all_jobs_status( session, err );
		if( OK(err) )
			sleep( 2 );
	 }

	DEBUG(( "<- drmaa_wait_thread; rc=%d: %s", err->rc, err->msg ));
	return NULL;
}



int
drmaa_wifexited( int *exited, int stat, char *errmsg, size_t errlen )
{
	*exited = ( stat < 128 );
	return DRMAA_ERRNO_SUCCESS;
}

int
drmaa_wexitstatus( int *exit_status, int stat, char *errmsg, size_t errlen )
{
	*exit_status = stat & 0x7f;
	return DRMAA_ERRNO_SUCCESS;
}

int
drmaa_wifsignaled( int *signaled, int stat, char *errmsg, size_t errlen )
{
	*signaled = ( stat >= 128 );
	return DRMAA_ERRNO_SUCCESS;
}

int
drmaa_wtermsig( char *signal, size_t signal_len, int stat,
		char *errmsg, size_t errlen )
{
	int sig = stat & 0x7f; /* stat - 128 */
#define DETECT_SIG( signame ) \
	case signame:  strlcpy( signal, #signame, signal_len );  break;
	switch( sig )
	 {
		/* signals described by POSIX.1 */
		DETECT_SIG( SIGHUP  )
		DETECT_SIG( SIGINT	)
		DETECT_SIG( SIGQUIT )
		DETECT_SIG( SIGILL	)
		DETECT_SIG( SIGABRT )
		DETECT_SIG( SIGFPE	)
		DETECT_SIG( SIGKILL )
		DETECT_SIG( SIGSEGV )
		DETECT_SIG( SIGPIPE )
		DETECT_SIG( SIGALRM )
		DETECT_SIG( SIGTERM )
		DETECT_SIG( SIGUSR1 )
		DETECT_SIG( SIGUSR2 )
		DETECT_SIG( SIGCHLD )
		DETECT_SIG( SIGCONT )
		DETECT_SIG( SIGSTOP )
		DETECT_SIG( SIGTSTP )
		DETECT_SIG( SIGTTIN )
		DETECT_SIG( SIGTTOU )
		/* signals described in SUSv2 and SUSv3 / POSIX 1003.1-2001 */
#ifdef SIGBUS
		DETECT_SIG( SIGBUS    )
#endif
#ifdef SIGPOLL
		DETECT_SIG( SIGPOLL	  )
#endif
#ifdef SIGPROF
		DETECT_SIG( SIGPROF	  )
#endif
#ifdef SIGSYS
		DETECT_SIG( SIGSYS	  )
#endif
#ifdef SIGTRAP
		DETECT_SIG( SIGTRAP	  )
#endif
#ifdef SIGURG
		DETECT_SIG( SIGURG	  )
#endif
#ifdef SIGVTALRM
		DETECT_SIG( SIGVTALRM )
#endif
#ifdef SIGXCPU
		DETECT_SIG( SIGXCPU	  )
#endif
#ifdef SIGXFSZ
		DETECT_SIG( SIGXFSZ	  )
#endif
		default:
			snprintf( signal, signal_len, "unknown signal=%d", sig );
			break;
	 }
	return DRMAA_ERRNO_SUCCESS;
}


int
drmaa_wcoredump( int *core_dumped, int stat, char *errmsg, size_t errlen )
{
	/* We have no way to find out whether a core file was created. */
	*core_dumped = 0;
	return DRMAA_ERRNO_SUCCESS;
}


int
drmaa_wifaborted( int *aborted, int stat, char *errmsg, size_t errlen )
{
	if( stat < 128 )
		*aborted = false;
	else switch( stat & 0x7f )
	 {
		case SIGTERM:  case SIGKILL:
			*aborted = true;
			break;
		default:
			*aborted = false;
			break;
	 }
	return DRMAA_ERRNO_SUCCESS;
}

