# -*- coding: UTF-8 -*-
#
# $Id: pbs_tests.py 533 2007-12-22 15:25:42Z lukasz $
#
# Testsuite for DRMAA library
# Copyright (C) 2006-2007  Fedstage Systems Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest
import os, os.path as osp

import drmaa
from drmaa import pbs_drmaa
from drmaa.testsuite.utils import *



class Configuration_valid( DRMAA_test_case ):
	def test_minimal( self ):
		pbs_drmaa.read_configuration( "" )

	def test_typical( self ):
		pbs_drmaa.read_configuration( r"""
# sample configuration
pool_delay: 10,
wait_thread: 1,
job_categories: {
	longterm: "-p -100 -l nice=10",
	amd64: "-l arch=amd64",
	python: "-l software=python25",
	java: "-l software=java,vmem=500mb",
},
""" )

	def test_complex( self ):
		pbs_drmaa.read_configuration( r"""{
foo: "bar", # comment
baz
 :
value,
'a key': { @:-1,_empty:{},
	one: {*:1}, },
job_categories: { default: "" }
}
""" )


class Configuration_invalid( DRMAA_test_case ):
	@drmaa_xfail( drmaa.ERRNO_INTERNAL_ERROR )
	def invalid_conf( self, conf ):
		pbs_drmaa.read_configuration( conf )
			
	def test_invalid_syntax( self ):
		self.invalid_conf( "," )
		self.invalid_conf( "{ foo:1 " )
		self.invalid_conf( "{ 1: value }" )
	
	def test_invalid_option_type( self ):
		self.invalid_conf( "wait_thread: string" )
		self.invalid_conf( "pool_delay: {}" )
		self.invalid_conf( "pool_delay: -1" )
		self.invalid_conf( "job_categories: 1" )


class Configuration_merge_job_categories( DRMAA_test_case ):
	def setUp( self ):
		DRMAA_test_case.setUp( self )
		pbs_drmaa.read_configuration(
				"job_categories: { A: '-v TEST=1', C: '-v TEST=1' }" )
		pbs_drmaa.read_configuration(
				"job_categories: { B: '-v TEST=2', C: '-v TEST=2' }" )

	def test_a( self ):
		self.assertEquals(
				self.environ( job_category="A" )["TEST"],	"1" )
	def test_b( self ):
		self.assertEquals(
				self.environ( job_category="B" )["TEST"],	"2" )
	def test_c( self ):
		self.assertEquals(
				self.environ( job_category="C" )["TEST"],	"2" )

	#def test_merge_categories( self ):
	#	for category in "ABC":
	#		env = self.environ( job_category=category )
	#		self.assertEquals( env["TEST"], dict(A="1", B="2", C="2")[category] )



class Native_spec_merging( DRMAA_test_case ):
	def test_env( self ):
		pbs_drmaa.read_configuration( r"""
job_categories: {
	default: '-v CATAGORY=default,FROM=default_category',
	a: '-v CATAGORY=a,FROM=a_category',
}
""" )
		def strip( env ):
			return dict(
					filter( lambda x: x[0] in ["CATAGORY","NATIVE","ATTRIB"],#"FROM"],
						env.iteritems() ) )
		def check( attrib, category, native ):
			job_attributes = {}
			valid = {}
			if category:
				job_attributes[ "job_category" ] = category
				valid["CATAGORY"] = category
				#valid["FROM"] = "%s_category" % category
			else:
				valid["CATAGORY"] = "default"
				#valid["FROM"] = "default_category"
			if attrib:
				job_attributes[ "v_env" ] = ["ATTRIB=1,FROM=attrib"]
				valid["ATTRIB" ] = "1"
				#valid["FROM"] = "attrib"
			if native:
				job_attributes[ "native_specification" ] = "-v NATIVE=1,FROM=native"
				valid["NATIVE"] = "1"
				#valid["FROM"] = "native"
			env = self.environ( **job_attributes )
			env = strip( env )
			self.assertEquals( env, valid )

		for attrib in (False, True):
			for category in (False, "default", "a"):
				for native in (False, True):
					check( attrib, category, native )



class Native_test( DRMAA_test_case ):
	def test_user( self ):
		jt = self.session.job_template( "sleep", 5, \
				job_name="drmaa.testsuite", native_specification="-u nosuchuser"
				)
		drmaa_xfail( lambda: jt.run(), drmaa.ERRNO_INVALID_ATTRIBUTE_VALUE )

	def test_nice( self ):
		jt = self.session.job_template( "sleep", 5, \
				job_name="drmaa.testsuite", native_specification="-l nice=10"
				)
		jt.run().wait()
	
	def test_queue( self ):
		jt = self.session.job_template( "sleep", 5, \
				job_name="drmaa.testsuite", native_specification="-q notexistent"
				)
		drmaa_xfail( lambda: jt.run(), drmaa.ERRNO_INVALID_ATTRIBUTE_VALUE )
	
	def test_hold_types( self ):
		jt = self.session.job_template( "sleep", 5, \
				job_name="drmaa.testsuite", native_specification="-h"
				)
		job = jt.run()
		job.wait_for_state( "user_on_hold" )
		job.release()
		job.wait_for_state( "running", "user_on_hold" )
		job.wait()
	
	def test_many( self ):
		job = self.sleep( native_specification="-h -l nice=10" )
		job.wait_for_state( "user_on_hold" )
		job.release()
		job.wait()


class Native_test_long_options( DRMAA_test_case ):
	def test_valid( self ):
		natives = [
			"--stagein",
			"--stagein=",
			"--stageout=",
			"--depend",
		]
		for native in natives:
			self.sleep( native_specification=native ).wait()
	
	def test_valid_user( self ):
		import pwd
		users = set( [ os.getuid(), os.geteuid() ] )
		for user in users:
			job = self.sleep(
					native_specification="-u %s" % pwd.getpwuid(user).pw_name )
			job.wait()

	def test_valid_group( self ):
		import grp
		groups = set( [ os.getgid(), os.getegid() ] + os.getgroups() )
		for group in groups:
			job = self.sleep(
					native_specification="--group-list=%s" \
					% grp.getgrgid(group).gr_name )
			job.wait()

	def test_syntax_error( self ):
		natives = [
			"--foo",
			"--", "-- ",
			"--=",
			"--stageinx bar",
			"--stageinx=baz",
		]
		for native in natives:
			jt = self.session.job_template( "sleep", 2, \
					native_specification=native, job_name="drmaa.testsuite" )
			drmaa_xfail( lambda: jt.run(), drmaa.ERRNO_INVALID_ATTRIBUTE_VALUE,
					"for native specification %r" % native )

	def test_invalid_value( self ):
		natives = [
			"--depend=bar",
			"--group-list=notexistent",
		]
		for native in natives:
			jt = self.session.job_template( "sleep", 2, \
					native_specification=native, job_name="drmaa.testsuite" )
			drmaa_xfail( lambda: jt.run(), drmaa.ERRNO_INVALID_ATTRIBUTE_VALUE,
					"for native specification %r" % native )



