# -*- coding: UTF-8 -*-
#
# $Id: utils.py 533 2007-12-22 15:25:42Z lukasz $
#
# Testsuite for DRMAA library
# Copyright (C) 2006-2007  Fedstage Systems Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os, os.path as osp
import unittest
import traceback

import drmaa


class DRMAA_test_case( unittest.TestCase ):
	def __init__( self, *args, **kw ):
		unittest.TestCase.__init__( self, *args, **kw )
		self.basedir = osp.dirname( osp.abspath(
					__file__.replace("build/lib/","") ) )
		for filename in ["sleeper_job", "exit_job", "suicide_job", "environ_job"]:
			setattr( self, filename, osp.join( self.basedir, "jobs", filename ) )

	def setUp( self ):
		self.session = drmaa.Session()

	def tearDown( self ):
		self.session.exit()
		del self.session

	def job_template( self, *argv, **attribs ):
		attribs.setdefault( "job_name", "drmaa.testsuite" )
		return self.session.job_template( *argv, **attribs )
	
	def job( self, *argv, **attribs ):
		return self.job_template( *argv, **attribs ).run()

	def sleep( self, seconds=2, **attribs ):
		return self.job( "sleep", seconds, **attribs )

	def output( self, *argv, **attribs ):
		output_file = osp.join( self.basedir, "test.output" )
		if os.access( output_file, os.F_OK ):
			os.remove( output_file )
		job = self.job( output_path=output_file, \
				*argv, **attribs )
		status = job.wait()
		assert status.get("exited",False) and status.get("exitstatus",1) == 0
		result = open( output_file, "rb" ).read()
		os.remove( output_file )
		return result

	def environ( self, **attribs ):
		lines = self.output( self.environ_job, **attribs )
		env = {}
		for line in lines.split("\n"):
			parts = line.split("=")
			if len(parts) > 1:
				name = parts[0]
				value = "=".join( parts[1:] )
				env[ name ] = value
		return env



class Exception_chain( Exception ):
	def __init__( self, *args ):
		import sys
		Exception.__init__( self, *args )
		self.chain = [ ( Exception_chain, self, traceback.extract_stack()[:-1] ) ]
		exc_type, exception, stacktrace = sys.exc_info()
		if hasattr( exception, "chain" ):
			self.chain += exception.chain
		else:
			self.chain.append(
					(exc_type, exception, traceback.extract_tb(stacktrace)) )
	def __str__( self ):
		return self.format()
	def format( self ):
		rep = "Traceback (most recent call last):\n"
		first = True
		for exc_type, exc, stack in self.chain:
			if not first:
				rep += "caused by:\n"
			for frame in traceback.format_list( stack ):
				rep += frame
			if hasattr( exc, "chain" ):
				message = Exception.__str__( exc )
			else:
				message = str(exc)
			type = str(exc_type).split(".")[-1]
			rep += "%s: %s\n" % (type, message)
			first = False
		return rep



def drmaa_xfail( *args ):
	function = None
	message = None
	if len(args) == 3:
		function, error_code, message = args
	elif len(args) == 2:
		function, error_code = args
	elif len(args) == 1: # called as decorator
		function, error_code = None, args[0]
	else:
		raise TypeError()

	if message is not None:
		tail = "; %s" % message
	else:
		tail = ""

	def decorator( f ):
		def wrapper( *args, **kwargs ):
			try:
				result = f( *args, **kwargs )
				assert False, "expected to fail with code %d but passed%s" \
					% (error_code, tail)
			except drmaa.drmaa_error, e:
				assert e.code == error_code, \
					"expected to fail with code %(expected)d but %(returned)d returned%(tail)s" \
					% dict( expected=error_code, returned=e.code, tail=tail )
		return wrapper

	if function is None: # act as an dectorator
		return decorator
	else: # call in place
		return decorator( function )()
	

def construct_job( session, argv ):
	jt = drmaa.Job_template( session )
	jt.job_name       = "drmaa.testsuite"
	jt.remote_command = argv[0]
	jt.v_argv         = argv[1:]
	return jt

def sleeper_job( session, time ):
	return session.job_template( "sleep", time, job_name="drmaa.testsuite" )
	#return construct_job( session, [ "sleep", str(time) ] )

